# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc


class ModelWithFile(abc.ABC):
    def __init__(self):

        super().__init__()

    @classmethod
    @abc.abstractmethod
    def get_model_file(cls) -> str:
        pass
